# math.py
# -*- coding: utf-8 -*-

from typing import Tuple

import particlepy.particle


def fade_color(particle: particlepy.particle.Particle, color: Tuple[int, int, int], progress: float):
    """Fades color of :attr:`particle` over life span (:attr:`progress`) to new color (:attr:`color`)

    Args:
        particle (:class:`particlepy.particle.Particle`): Particle to fade color with
        color (Tuple[int, int, int]): Color to fade to
        progress (float): Life span identifier:
            :attr:`particlepy.particle.Particle.progress` or :attr:`particlepy.particle.Particle.inverted_progress`

    Returns:
        List[float]: New color of particle
    """
    return [particle.shape.start_color[0] + (color[0] - particle.shape.start_color[0]) * progress,
            particle.shape.start_color[1] + (color[1] - particle.shape.start_color[1]) * progress,
            particle.shape.start_color[2] + (color[2] - particle.shape.start_color[2]) * progress]


def fade_alpha(particle: particlepy.particle.Particle, alpha: int, progress: float):
    """Fades :attr:`alpha` (transparency) of argument :attr:`particle` over life span (:attr:`progress`) to new color (:attr:`color`)

    Args:
        particle (:class:`particlepy.particle.Particle`): Particle to alpha color with
        alpha (int): Transparency to fade to
        progress (float): Life span identifier:
            :attr:`particlepy.particle.Particle.progress` or :attr:`particlepy.particle.Particle.inverted_progress`

    Returns:
        float: New alpha of particle
    """
    return particle.shape.start_alpha + (alpha - particle.shape.start_alpha) * progress



# shape.py
# -*- coding: utf-8 -*-

from typing import Tuple
from abc import ABC
import contextlib

with contextlib.redirect_stdout(None):
    import pygame


# TODO: AA shapes


class BaseShape(object):
    """The basic shape class. Is used as :attr:`shape` argument in :class:`particlepy.particle.Particle`.
        Is subclassed to create other shapes, e.g. :class:`Circle` or :class:`Rect`

    Args:
        radius (float): Radius of shape
        color (Tuple[int, int, int]): Color of shape
        alpha (int, optional): Transparency of shape `(0 - 255)`, defaults to `255`
        angle (float, optional): Degrees of rotation of shape, defaults to `0`

    Attributes:
        radius (float): Radius of shape
        _start_radius (float): Radius of shape when being instanced. Property is :func:`BaseShape.start_radius`
        angle (int): Degrees of rotation of shape
        color (List[int, int, int]): Color of shape
        _start_color (Tuple[int, int, int]): Color of shape when being instanced. Property is :func:`BaseShape.start_color`
        alpha (int): Transparency of shape, ranges from `0` to `255`
        _start_alpha (int): Transparency of shape when being instanced. Property is :func:`BaseShape.start_alpha`
        surface (:class:`pygame.Surface`): Pygame surface of shape
        rect (:class:`pygame.Rect`): Pygame Rect of :attr:`surface`. Position does not affect anything
    """

    def __init__(self, radius: float, color: Tuple[int, int, int], alpha: int = 255, angle: float = 0):
        self.radius = radius
        self._start_radius = self.radius

        self.angle = angle

        self.color = list(color)
        self._start_color = tuple(self.color)
        self.alpha = alpha
        self._start_alpha = self.alpha

        self.rect = None
        self.surface = self.make_surface()

    @property
    def start_radius(self):
        """Returns :attr:`_start_radius`

        Returns:
            float: :attr:`_start_radius`
        """
        return self._start_radius

    @property
    def start_color(self):
        """Returns :attr:`_start_color`

        Returns:
            Tuple[int]: :attr:`_start_color`
        """
        return self._start_color

    @property
    def start_alpha(self):
        """Returns :attr:`_start_alpha`

        Returns:
            int: :attr:`_start_alpha`
        """
        return self._start_alpha

    def decrease_radius(self, delta_radius: float):
        """Decreases radius of shape by :attr:`delta_radius`

        Args:
            delta_radius (float): Radius decrease value
        """
        self.radius -= delta_radius
        if self.radius < 0:
            self.radius = 0

    def make_surface(self) -> pygame.Surface:
        """Creates shape surface and rect by calling :func:`BaseShape.make_shape()` and :func:`BaseShape.rotate()`

        Returns:
            :class:`pygame.Surface`: Currently created shape surface (:attr:`surface`)
        """
        self.surface = pygame.Surface((self.radius * 2, self.radius * 2), pygame.SRCALPHA)
        self.surface.set_alpha(self.alpha)
        self.make_shape()
        self.rotate_shape()
        self.rect = self.surface.get_rect()
        return self.surface

    def make_shape(self):
        """Creates shape for shape surface. Can be modified to make different shapes and effects.
        """
        raise NotImplementedError("BaseShape.make_shape() creates no shape."
                                  "Own functions have to be written for custom shapes.")

    def rotate_shape(self):
        """Rotates shape. Only called by :func:`BaseShape.make_shape()`
        """
        if self.radius > 1:  # pygame issue: https://github.com/pygame/pygame/issues/2464
            self.surface = pygame.transform.rotate(self.surface, self.angle)


class Circle(BaseShape, ABC):
    """Circle shape class. Is subclass of :class:`BaseShape` and inherits all attributes and methods

    Args:
        radius (float): Radius of shape
        color (Tuple[int, int, int]): Color of shape
        alpha (int, optional): Transparency of shape `(0 - 255)`, defaults to `255`
        angle (float, optional): Degrees of rotation of shape, defaults to `0`

    Attributes:
        radius (float): Radius of shape
        _start_radius (float): Radius of shape when being instanced. Property is :func:`BaseShape.start_radius`
        angle (int): Degrees of rotation of shape
        color (List[int, int, int]): Color of shape
        _start_color (Tuple[int, int, int]): Color of shape when being instanced. Property is :func:`BaseShape.start_color`
        alpha (int): Transparency of shape, ranges from `0` to `255`
        _start_alpha (int): Transparency of shape when being instanced. Property is :func:`BaseShape.start_alpha`
        surface (:class:`pygame.Surface`): Pygame surface of shape
    """

    def __init__(self, radius: float, color: Tuple[int, int, int], alpha: int = 255, angle: float = 0):
        super(Circle, self).__init__(radius, color, alpha, angle)

    def make_shape(self):
        """Makes a circle
        """
        pygame.draw.circle(self.surface, self.color, (self.radius, self.radius), self.radius)


class Rect(BaseShape, ABC):
    """Rectangle shape class. Is subclass of :class:`BaseShape` and inherits all attributes and methods

    Args:
        radius (float): Radius of shape
        color (Tuple[int, int, int]): Color of shape
        alpha (int, optional): Transparency of shape `(0 - 255)`, defaults to `255`
        angle (float, optional): Degrees of rotation of shape, defaults to `0`

    Attributes:
        radius (float): Radius of shape
        _start_radius (float): Radius of shape when being instanced. Property is :func:`BaseShape.start_radius`
        angle (int): Degrees of rotation of shape
        color (List[int, int, int]): Color of shape
        _start_color (Tuple[int, int, int]): Color of shape when being instanced. Property is :func:`BaseShape.start_color`
        alpha (int): Transparency of shape, ranges from `0` to `255`
        _start_alpha (int): Transparency of shape when being instanced. Property  is :func:`BaseShape.start_alpha`
        surface (:class:`pygame.Surface`): Pygame surface of shape
    """

    def __init__(self, radius: float, color: Tuple[int, int, int], alpha: int = 255, angle: float = 0):
        super(Rect, self).__init__(radius, color, alpha, angle)

    def make_shape(self):
        """Makes a rectangle
        """
        self.surface.fill(self.color)



# particle.py
# -*- coding: utf-8 -*-

from typing import Tuple, List
import copy
import contextlib

with contextlib.redirect_stdout(None):
    import pygame

import particlepy.shape


class Particle(object):
    """This is the particle class. It simulates the physics of a particle and can be used in a particle system (:class:`ParticleSystem`)

    Args:
        shape (:class:`particlepy.shape.BaseShape`): Visual particle shape
        position (Tuple[float, float]): Center position
        velocity (Tuple[float, float]): Velocity
        delta_radius (float): Radius decrease value
        data (dict, optional): A dictionary for extra data, defaults to `None`
        alive (bool, optional): `True` if particle should be alive, and `False` if otherwise, defaults to `True`

    Attributes:
        shape (:class:`particlepy.shape.BaseShape`): Visual particle shape
        position (List[float, float]): Center position
        velocity (List[float, float]): Velocity (can be modified with gravity)
        delta_radius (float): Radius decrease value
        progress (float): A variable ranging from 0 to 1 to represent the lifespan
        inverted_progress (float): A variable ranging from 1 to 0 to represent the lifespan
        time (float): A simple timer
        data (dict): A dictionary for extra data
        alive (bool): `True` if particle is alive, and `False` if otherwise
    """

    def __init__(self, shape: particlepy.shape.BaseShape, position: Tuple[float, float], velocity: Tuple[float, float],
                 delta_radius: float, data: dict = None, alive: bool = True):
        """Constructor method
        """
        self.shape = copy.copy(shape)

        self.position = list(position)
        self.velocity = list(velocity)

        self.delta_radius = delta_radius

        self.progress, self.inverted_progress = 0, 1
        self.get_progress()

        self.time = 0

        if data:
            self.data = data
        else:
            self.data = {}

        self.alive = alive

    def kill(self):
        """Sets attribute :attr:`alive` `False`
        """
        self.alive = False

    def revive(self):
        """Sets attribute :attr:`alive` `True`
        """
        self.alive = True

    def get_progress(self) -> Tuple[float, float]:
        """Returns tuple of two floats: `progress` and `inverted_progress`

        Returns:
            Tuple[float, float]: `progress` and `inverted_progress`
        """
        progress = self.shape.radius / self.shape.start_radius
        self.progress, self.inverted_progress = progress, 1 - progress
        return self.progress, self.inverted_progress

    def update(self, delta_time: float, gravity: Tuple[float, float] = None):
        """Updates position, velocity, progress, etc. of particle and kills it, if :code:`radius <= 0`

        Args:
            delta_time (float): A value to let the particle move according to frame time
            gravity (Tuple[float, float], optional): Affects the velocity and 'pulls' it in a direction, defaults to None
        """
        self.shape.decrease_radius(self.delta_radius)
        if self.shape.radius > 0:
            if self.alive:
                self.position[0] += self.velocity[0] * delta_time
                self.position[1] += self.velocity[1] * delta_time
                if gravity:
                    self.velocity[0] += gravity[0]
                    self.velocity[1] += gravity[1]

                self.get_progress()
                self.time += delta_time
        else:
            self.kill()

    def render(self, surface: pygame.Surface):
        """Renders the particle on given surface

        Args:
            surface (:class:`pygame.Surface`): The surface on which the particle is being rendered on
        """
        if self.alive:
            surface.blit(self.shape.surface, (self.position[0] - self.shape.surface.get_width() / 2,
                                              self.position[1] - self.shape.surface.get_height() / 2))


class ParticleSystem(object):
    """The particle system class. It is used to manage particles in a group

    Args:
        data (dict, optional): A dictionary for extra data, defaults to None
        alive (bool, optional): `True` if particle system should be alive, and `False` if otherwise, defaults to `True`

    Attributes:
        particles (List[:class:`Particle`])
        data (dict): A dictionary for extra data
        alive (bool): `True` if particle system is alive, and `False` if otherwise
    """

    def __init__(self, data: dict = None, alive: bool = True):
        """Constructor method
        """
        self.particles: List[particlepy.particle.Particle] = []
        if data:
            self.data = data
        else:
            self.data = {}
        self.alive = alive

    def emit(self, particle: Particle):
        """Creates a new particle

        Args:
            particle (:class:`Particle`): Particle which is being created

        Raises:
            Exception: Particle system is not alive, not able to add particles
        """
        if self.alive:
            self.particles.append(copy.copy(particle))
        else:
            raise Exception("Particle system is not alive, not able to add particles")

    def clear(self):
        """Clears the particle list
        """
        self.particles.clear()

    def kill(self):
        """Sets :attr:`alive` `False`
        """
        self.alive = False

    def revive(self):
        """Sets :attr:`alive` `True`
        """
        self.alive = True

    def update(self, delta_time: float, gravity: Tuple[float, float] = None):
        """Calls :func:`Particle.update()` for every particle in system

        Args:
            delta_time (float): A value to let the particles move according to frame time
            gravity (Tuple[float, float], optional): Affects the velocity and 'pulls' particles in a direction, defaults to None
        """
        if self.alive:
            for particle in self.particles:
                particle.update(gravity=gravity, delta_time=delta_time)
                if not particle.alive:
                    self.particles.remove(particle)

    def make_shape(self):
        """Makes the surface of all particles in system
        """
        if self.alive:
            for particle in self.particles:
                particle.shape.make_surface()

    def render(self, surface: pygame.Surface):
        """Renders surface of all particles on given surface

        Args:
            surface (:class:`pygame.Surface`): Surface on which the particles are being rendered
        """
        if self.alive:
            for particle in self.particles:
                particle.render(surface=surface)
